;-- C64 Emulator
.MODEL Large,PASCAL
.386

;==========================================================
BEGINP MACRO Name
 Align 2
 Name PROC NEAR
ENDM

LED Macro Color
 MOV DX,03C4h
 MOV AX,0F02h
 OUT DX,AX
 MOV AL,Color
 MOV ES,VGASeg
 MOV DI,VGAOfs
 MOV ES:[DI+5+222*90],AL
 MOV ES:[DI+5+223*90],AL
 MOV ES:[DI+5+224*90],AL
ENDM

CALLP MACRO Name,LED,Color
 PUSHA

 MOV ES,Seg0040
 MOV EAX,ES:[6Ch]
 MOV NoTicks,EAX

 &LED &Color

 CALL Name

 MOV ES,Seg0040
 MOV EAX,ES:[6Ch]
 SUB EAX,NoTicks
 SUB Ticks,EAX

 MOV ES,ROMSeg
 POPA
ENDM

 rEBX = 5Dh
 rEDX = 55h

JMP32 Macro Reg,Base  ; BP can't link an obj with 32bit segments :(
 db 67h,0FFh,24h  ; JMP [2*
 db r&Reg         ; REG32+
 dw Offset Base,0 ; longint(base)]
ENDM

 ROM equ ES
 RAM equ FS
 CIO equ GS

;==========================================================
.DATA

 GNU db 'CBM 64 Emulator v0.2m',13,10
     db 'Copyright (C) 1996-98  Paul Toth (tothpaul@mygale.org)',13,10
     db 'http://www.mygale.org/~tothpaul',13,10
     db 10
     db ' This program is free software; you can redistribute it and/or modify',13,10
     db 'it under the terms of the GNU General Public License as published by',13,10
     db 'the Free Software Foundation; either version 2 of the License, or',13,10
     db '(at your option) any later version.',13,10
     db 10
     db ' This program is distributed in the hope that it will be useful,',13,10
     db 'but WITHOUT ANY WARRANTY; without even the implied warranty of',13,10
     db 'MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the',13,10
     db 'GNU General Public License for more details.',13,10
     db 10
     db ' You should have received a copy of the GNU General Public License',13,10
     db 'along with this program; if not, write to the Free Software',13,10
     db 'Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.',13,10
     db 10,'$'

IF DEBUG
 EXTRN dbgCPU:FAR
 EXTRN dbgVIC:FAR
 EXTRN dbgNMI:FAR
 EXTRN dbgIRQ:FAR
 EXTRN dbgBRK:FAR
 EXTRN dbgPCRAM:FAR
 EXTRN dbgPCROM:FAR
 EXTRN dbgDRAM:FAR
 EXTRN dbgDROM:FAR
 EXTRN dbgDCIO:FAR
 EXTRN dbgCFG:FAR
 EXTRN dbgKON:FAR
 EXTRN dbgKOFF:FAR
 EXTRN dbgBON:FAR
 EXTRN dbgBOFF:FAR
 EXTRN dbgPokeDC0D:FAR
 EXTRN dbgOutput:FAR
 EXTRN dbgOutputATN:FAR
 EXTRN dbgOutputSEC:FAR
 EXTRN dbgInput:FAR
 EXTRN dbgVICIRQ:FAR
 EXTRN dbgVICBase:FAR
 EXTRN dbgVICMode:FAR
 EXTRN dbgDx16Before:FAR
 EXTRN dbgDx16After :FAR
 EXTRN dbgKEYB:FAR
ELSE
 EXTRN PAUSE:FAR           ;-- Pause Key
 EXTRN StatusInfo:FAR      ;-- F10
 EXTRN Joy1Active:FAR
 EXTRN Joy2Active:FAR
ENDIF

 EXTRN ChangeDisk:FAR      ;-- F9=ChangeDisk
 EXTRN CPUJAM:FAR

 EXTRN DOSErr:byte	   ;-- for GetMem

 EXTRN RAMSeg :WORD	   ;-- Segment for C64
 EXTRN ROMSeg :WORD
 EXTRN CIOSeg :WORD
 EXTRN Seg0040:WORD	   ;-- Segment for BIOS Ticks

 EXTRN BASIC_ROM  :WORD    ;-- ROMs
 EXTRN CHARSET_ROM:WORD
 EXTRN KERNEL_ROM :WORD

 EXTRN Ticks :DWORD	   ;-- Timming
 EXTRN Cycles:WORD
 EXTRN TotalCycles:DWORD
 EXTRN Rate :WORD

 EXTRN PC:WORD		   ;-- C64 Registers
 EXTRN A :WORD
 EXTRN X :WORD
 EXTRN Y :WORD
 EXTRN S :WORD
 EXTRN P :BYTE

 EXTRN opJAM:BYTE

;==========================================================
;--- Flags

 CFlag equ 1
 ZFlag equ 2
 IFlag equ 4
 DFlag equ 8
 BFlag equ 16
 _Flag equ 32
 VFlag equ 64
 NFlag equ 128

;==========================================================

INCLUDE VIC2.ASM
INCLUDE PEEK.ASM
INCLUDE POKE.ASM
INCLUDE CIA1.ASM
INCLUDE CIA2.ASM
INCLUDE 6510.ASM
IFE DEBUG
 INCLUDE KEYB.ASM
 INCLUDE VGAX.ASM
ENDIF

 NoTicks  dd 0

;==========================================================
.CODE

 PUBLIC Startup
 PUBLIC Emulate

;==========================================================
Startup PROC NEAR

;-- Copy ROMs

 MOV ES,ROMSeg

 MOV SI,offset BASIC_ROM
 MOV DI,0A000h	; BASIC ROM
 MOV CX,8192/4
 REP MOVSD

 MOV SI,offset CHARSET_ROM
 MOV DI,0D000h	; CHARS ROM
 MOV CX,4096/4
 REP MOVSD

 MOV SI,offset KERNEL_ROM
 MOV DI,0E000h	; KERNAL ROM
 MOV CX,8192/4
 REP MOVSD

;-- Patch Kernal (Frodo)

 MOV ES,ROMSeg

;-- LOAD DISK (.D64)

; OutputATN (Talk/Unlisten/Untalk) on serial bus
 MOV WORD PTR ES:[0ED23h],01F2h

; Output byte from $95 on serial bus
 MOV WORD PTR ES:[0ED40h],00F2h

; OutputSEC from $95 on serial bus
 MOV WORD PTR ES:[0ED36h],02F2h

; Input on serial bus
 MOV WORD PTR ES:[0EE13h],03F2h

; Only needed for real IEC
 MOV BYTE PTR ES:[0EDEFh],60h
 MOV BYTE PTR ES:[0EDBEh],60h
 MOV BYTE PTR ES:[0EDCCh],60h
 MOV BYTE PTR ES:[0EE03h],60h

;-- LOAD TAPE (.T64)
; Read TAPE Block
 MOV WORD PTR ES:[0F841h],04F2h


; Patch Default device 1 -> 8
 MOV BYTE PTR ES:[0E1DAh],08h ; E1D9 LDX $01 ; default = tape

Reset:
 Call ResetCIA1

 MOV ES,RAMSeg  ; Clear RAM
 XOR EAX,EAX
 MOV DI,AX
 MOV CX,0FFFFh/4
 REP STOSD

 MOV ES,CIOSeg  ; Clear CHIPS
 MOV DI,0D000h
 XOR EAX,EAX
 MOV CX,4096/4
 REP STOSD

; Init Regs
 MOV PC,0FCE2h
 MOV P,_Flag
 MOV S,1FFh
 MOV opJAM,0

 MOV CIA1,0 ; No IRQ
 MOV CIA2,0 ; No NMI

 XOR ECX,ECX
 MOV Cycles,CX
 MOV TotalCycles,ECX

 ;-- for Reset (F11 or JAM)
 MOV RAM,RAMSeg
 MOV ROM,ROMSeg
 MOV CIO,CIOSeg

 RET
Startup ENDP

;==========================================================

Emulate PROC NEAR

IFE DEBUG
 Call InstallKeyboard
 Call TweakMode
ENDIF

 MOV ES,Seg0040
 MOV EAX,ES:[6Ch]
 MOV Ticks,EAX

 MOV RAM,RAMSeg
 MOV ROM,ROMSeg
 MOV CIO,CIOSeg

Align 2
Next:

 SUB Cycles,CX
 JNS SkipVic

  ADD TotalCycles,64  ; ??

  Call UpdateRaster   ; Inc(Raster) and Check IRQRaster
  Call DrawLine       ; for Current Raster
  Call CIA1Line       ; Timers
  Call CIA2Line       ; Timers, IRQ & NMI

 SkipVic:

IF DEBUG
 callp dbgCPU
ENDIF


ExitPatch:
 JMP PeekPC ;-- goto OpTable[OpCode] (6510.ASM)
 JMP Done   ;-- exit patched (KEYB.ASM)

;==========================================================
;--- Text Mode ---

JAM:
 MOV opJAM,DL
 CALLP CPUJAM
 Call Reset
 Call NewBank
 Call NewVideoMode
 JMP  ConfigMem ; Restore ROMS !!!
Done:

IFE DEBUG
 Call RemoveKeyboard
 Call TextMode
 MOV  DX,Offset GNU
 MOV  AH,09h
 INT  21h
ENDIF

;--- Consumme Ticks
 MOV ES,Seg0040
 MOV EAX,ES:[6Ch]
 SUB EAX,Ticks
 MOV EBX,10
 MUL EBX
 MOV EBX,182 ; /18.2 ...
 DIV EBX
 MOV Ticks,EAX
 JZ  ZeroRate
 MOV EBX,EAX
 MOV EAX,TotalCycles
 XOR EDX,EDX
 DIV EBX
 XOR EDX,EDX
 MOV EBX,10000
 DIV EBX
 ZeroRate:
 MOV Rate,AX

 RET

Emulate ENDP

END
