;--- CIA2 Emulation

;===============================================================
.DATA

	   ;  DDx0  DDx1  DDx2  DDx3  DDx4  DDx5  DDx6  DDx7
PokeCIA2  dw wDDx0,wDDxx,wDDx0,wDDxx,wDDxx,wDDx5,wDDxx,wDDx7
	   ;  DDx8  DDx9  DDxA  DDxB  DDxC  DDxD  DDxE  DDxF
	  dw wDDx8,wDDx9,wDDx9,wDDx9,wDDxC,wDDxD,wDDxE,wDDxF

	   ;  DDx0  DDx1  DDx2  DDx3  DDx4  DDx5  DDx6  DDx7
PeekCIA2  dw rDDx0,rDDx0,rDDxx,rDDxx,rDDx4,rDDx5,rDDx6,rDDx7
	   ;  DDx8  DDx9  DDxA  DDxB  DDxC  DDxD  DDxE  DDxF
	  dw rDDxx,rDDxx,rDDxx,rDDxx,rDDxx,rDDxD,rDDxx,rDDxx

Start2	dw 0  ; 1 = TimerA , 2 = TimerB , 4 = Chain Timers
TimerA2 dw 0
TimerB2 dw 0

CIA2  db 0    ; 1 = NMI
Mask2 db 0

NMIFlag = 1

Alarm2  dd 0   ; DDx8..DDxB when (DDxF and 80h>0)
TodDiv2 db 0  ; CountTOD is called by VBlank (should be 50Hz)

;===============================================================
.CODE

BEGINP CIA2Line

 MOV CX,Cycles ; Just set by VIC.DrawLine...

;-- Timer A
 TEST Start2,1
 JZ SkipTimerA2

  SUB TimerA2,CX     ; Timer A
  JNS SkipTimerA2

  MOV AX,CIO:[0DD04h] ; Reset Timer
  MOV TimerA2,AX

  MOV  AL,CIO:[0DD0Eh] ; Control Register A
  TEST AL,8
  JZ ContinueA2
   AND AL,0FEh
   MOV CIO:[0DD0Eh],AL
   AND Start2,2+4
  ContinueA2:

  OR byte ptr CIO:[0DD0Dh],1   ; IRQ Control Register
  TEST Mask2,1
  JZ NoNMIA
   OR byte ptr CIO:[0DC0Dh],80h
   OR CIA2,NMIFlag
  NoNMIA:

  TEST Start2,4
  JZ SkipTimerA2

  MOV CX,1
  JMP DoTimerB2

 SkipTimerA2:

;-- Timer B
 TEST Start2,2
 JZ SkipTimerB2

 DoTimerB2:

  SUB TimerB2,CX
  JNS SkipTimerB2

  MOV AX,CIO:[0DD06h]
  MOV TimerB2,AX

  MOV  AL,CIO:[0DD0Fh] ; Control Register
  TEST AL,8
  JZ ContinueB2
   AND AL,0FEh
   MOV CIO:[0DD0Fh],AL
   AND Start2,1
  ContinueB2:

  OR byte ptr CIO:[0DD0Dh],2   ; IRQ Control Register
  TEST Mask2,2
  JZ NoNMIB
   OR byte ptr CIO:[0DD0Dh],80h
   OR CIA2,NMIFlag
  NoNMIB:

 SkipTimerB2:

CIA2Line ENDP

BEGINP CheckNMI

 TEST CIA2,NMIFlag
 JNZ  ExecNMI

  JMP CheckIRQ ; CIA1.ASM

 ExecNMI:

  IF DEBUG
   CALLP dbgNMI
  ENDIF

  AND CIA2,255-NMIFlag
  MOV BX,S
  MOV AX,PC

  MOV RAM:[BX],AH
  DEC BL
  MOV RAM:[BX],AL
  DEC BL

  MOV AL,P

  MOV RAM:[BX],AL
  DEC BL

  MOV S,BX
  OR P,IFlag
  MOV BX,0FFFAh
  Call DeekBX
  MOV PC,BX
  SUB Cycles,7

 RET
CheckNMI ENDP

;===============================================================

BEGINP CountTOD2
 TEST Start2,8 ; CountTOD
 JZ   NoTOD2

 DEC TodDiv2
 JS  ProceedTod2
  RET
 ProceedTod2:
  MOV TodDiv2,4
  TEST BYTE PTR CIO:[0DD0Eh],80h ; CR A
  JNZ TodDiv2Ok
   INC TodDiv2
  TodDiv2Ok:

  MOV EAX,CIO:[0DD08h] ; TOD
  Call IncTime ; CIA1.ASM
  MOV CIO:[0DD08h],EAX

  CMP EAX,Alarm2
  JNE NoTOD2

   OR byte ptr CIO:[0DD0Dh],4   ; IRQ Control Register
   TEST Mask2,4
   JZ NoTOD2
    OR byte ptr CIO:[0DD0Dh],80h
    OR CIA1,4 ; NMI

  NoTOD2:
 RET
CountTOD2 ENDP


; POKE CIA2(DI),AL --------------------------

BEGINP wDDxx
 MOV CIO:[DI],AL
 JMP Next
wDDxx ENDP

BEGINP wDDx0     ; DDx0 & DDx2 (?)
 MOV CIO:[DI],AL
 Call NewBank    ; VIC2.ASM -> Change Video Bank
 JMP Next
wDDx0 ENDP

BEGINP wDDx5     ; Timer A (HighByte)
 MOV CIO:[DI],AL
 TEST Start2,1
 JNZ SkipDDx5
  MOV AX,CIO:[DI-1]
  MOV TimerA2,AX
 SkipDDx5:
 JMP Next
wDDx5 ENDP

BEGINP wDDx7     ; Timer B (HighByte)
 MOV CIO:[DI],AL
 TEST Start2,2
 JNZ SkipDDx7
  MOV AX,CIO:[DI-1]
  MOV TimerB2,AX
 SkipDDx7:
 JMP Next
wDDx7 ENDP

BEGINP wDDx8 ;
 TEST BYTE PTR CIO:[0DC0Fh],80h
 JNZ  wDDx8_Alarm
  MOV CIO:[DI],AL
  OR  Start2,8 ; CountTOD
 JMP Next
wDDx8_Alarm:
 MOV BYTE PTR DS:[Alarm2+DI-0DD08h],AL
 JMP Next
wDDx8 ENDP

BEGINP wDDx9 ; ..xDCxB
 TEST BYTE PTR CIO:[0DD0Fh],80h
 JNZ   wDDx8_Alarm
  MOV CIO:[DI],AL
  AND Start2,255-8 ; CountTOD
 JMP Next
wDDx9 ENDP

BEGINP wDDxC
 MOV CIO:[DI],AL
 OR byte ptr CIO:[DI+1],4   ; IRQ Control Register
 TEST Mask2,4
 JZ wDDxC_done
  OR byte ptr CIO:[DI+1],80h
  OR CIA1,4 ; NMI
 wDDxC_done:
 JMP Next
wDDxC ENDP


BEGINP wDDxD
 TEST AL,128
 JNZ wDDxD_OrMask
  NOT AL
  AND AL,Mask2
  JMP wDDxD_DoMask
 wDDxD_OrMask:
  AND AL,7Fh
  OR  AL,Mask2

  MOV AH,CIO:[DI]
  AND AH,AL
  AND AH,1Fh
  JZ wDDxD_DoMask

   AND BYTE PTR CIO:[DI],80h
   OR  CIA1,4 ; NMI

 wDDxD_DoMask:
 MOV Mask2,AL

 JMP Next
wDDxD ENDP

BEGINP wDDxE
 TEST AL,10h
 JZ NotLoadA2
  MOV BX,CIO:[0DD04h]
  MOV TimerA2,BX
 NotLoadA2:
 MOV AH,AL
 AND AL,0EFh
 MOV CIO:[DI],AL
 AND AH,21h
 AND Start2,255-1
 CMP AH,1
 JNE NotStartA2
  OR Start2,1
 NotStartA2:
 JMP Next
wDDxE ENDP

BEGINP wDDxF
 TEST AL,10h
 JZ NotLoadB2
  MOV BX,CIO:[0DD06h]
  MOV TimerB2,AX
 NotLoadB2:
 MOV AH,AL
 AND AL,0EFh
 MOV CIO:[DI],AL
 AND AL,61h
 AND Start2,255-2-4
 CMP AH,1
 JNE NotStartB2
  OR Start2,2
 NotStartB2:
 CMP AH,41h
 JNE NotChain2
  OR Start2,4
 NotChain2:
 JMP Next
wDDxF ENDP

; AL=PEEK CIA2(DI) -------------------------

BEGINP rDDxx
 MOV AL,CIO:[DI]
 RET
rDDxx ENDP

BEGINP rDDx0     ; DDx0 & DDx1
 MOV AL,CIO:[DI+2]
 NOT AL
 OR  AL,CIO:[DI]
 RET
rDDx0 ENDP

BEGINP rDDx4
 MOV AL,BYTE PTR TimerA2
 RET
rDDx4 ENDP

BEGINP rDDx5
 MOV AL,BYTE PTR TimerA2+1
 RET
rDDx5 ENDP

BEGINP rDDx6
 MOV AL,BYTE PTR TimerB2
 RET
rDDx6 ENDP

BEGINP rDDx7
 MOV AL,BYTE PTR TimerB2+1
 RET
rDDx7 ENDP

BEGINP rDDxD
 MOV AL,CIO:[DI]
 MOV BYTE PTR CIO:[DI],0
 AND CIA1,255-4 ; CLEAR NMI
 RET
rDDxD ENDP

