Unit D64;

Interface

uses
 fdos,dialogs,keyboard,dos;

Type
{ BAM structure }
 TBAM=record
  Track :byte; { Directory }
  Sector:byte;
  Format:char; { 'A' }
  Zero  :byte; {  0  }
  SAT   :array[1..35] of longint; { Sector Allocation Table }
  Name  :array[1..18] of char;    { Label +  (160=RVS+32) }
  ID    :word; { '00'    }
  Space :byte; { ' '+128 }
  Chars :word; { '2A'    }
  Spaces:array[1..4] of byte;  { ' '+128 }
  Zeros :array[1..85] of byte; { 0 }
 end;

{ Directory entry structure }
 TDirEntry=record
  Kind   :byte;
  nTrack :byte;
  nSector:byte;
  Name   :array[1..16] of char;
  sTrack :byte;
  sSector:byte;
  Len    :byte;
  pad0   :longint;
  oTrack :byte;
  oSector:byte;
  Blocks :word;
  pad1   :word;
 end;

{ Directory block structure }
 TDirectory=record
  NextTrack :byte;
  NextSector:byte;
  Entry     :array[0..7] of TDirEntry;
 end;

 TSector=record
  NextTrack,NextSector:byte;
  Data:array[0..253] of byte;
 end;

{ TBuffer=array[0..8192] of byte; }

function OpenDevice:boolean;
function Load(Channel:byte; AFileName:string):byte;
function ReadByte(Channel:byte; Var Value:word):byte;
function Save(Channel:byte; AFileName:string):byte;
function WriteByte(Channel:byte; Value,EndOfFile:byte):byte;
procedure ChangeDisk;

Var
 D64File:string;

Implementation

uses iec;

const
 SectorCount:array[1..40] of byte=(
  21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,
  19,19,19,19,19,19,19,
  18,18,18,18,18,18,
  17,17,17,17,17,
  17,17,17,17,17	{ Tracks 36..40 }
 );

 SectorOfs:array[1..40] of integer=(
  0,21,42,63,84,105,126,147,168,189,210,231,252,273,294,315,336,
  357,376,395,414,433,452,471,
  490,508,526,544,562,580,
  598,615,632,649,666,
  683,700,717,734,751	{ Tracks 36..40 }
 );

 KindName:array[0..15] of array[1..3] of char=(
  'DEL','SEQ','PRG','USR','REL','EL?','EQ?','RG?','SR?','EL?','L??',
  'Q??','G??','R??','L??','???'
 );

Var
 Handle:word;

 HeaderSize:word;

 BAM:tBAM;
 FreeBlocks:word;

 FileName:string;

 Sector     :TSector;
 SectorSize :word;
 SectorIndex:word;
 NextSector :word;
 NextTrack  :word;

 Dir:TDirectory;
 DirTrack :word;
 DirSector:word;
 DirEntry :word;

 FillSector :(fsNone,fsRead,fsDir,fsDone);

Function SectorOffset(Track,Sector:word):longint;
 begin
  if (Track<1)or(Track>40)
   or(Sector>=SectorCount[Track]) then
    SectorOffset:=-1
   else
    SectorOffset:=(SectorOfs[Track]+Sector) shl 8+HeaderSize;
 end;

{ Read sector (256 bytes) }
function ReadSector(Track,Sector:byte; Var Data):boolean;
 var
  offset:longint;
 begin
{  StatusBar('Reading Sector'); }

  Offset:=SectorOffset(Track,Sector);
  ReadSector:=(Offset>=0)
           and(SeekFile(Handle,Offset))
           and(ReadFile(Handle,Data,256)=256);
 end;

function WriteSector(Track,Sector:byte; Var Data):boolean;
 var
  offset:longint;
 begin
{  StatusBar('Writing Sector'); }

  Offset:=SectorOffset(Track,Sector);
  WriteSector:=(Offset>=0)
            and(SeekFile(Handle,Offset))
            and(WriteFile(Handle,Data,256)=256);
 end;

Function AllocSector(Var ATrack,ASector:byte):boolean;
 Var
  T,S:byte;
  F:longint;
  A:longint;
 begin
  If FreeBlocks=0 then begin
{   StatusBar('Disk full'); }
   AllocSector:=False;
   exit;
  end;
  For T:=1 to 35 do begin { Search in all Tracks }
   F:=BAM.SAT[T]; { SAT info for this Track }
   if (F AND $FF)>0 then begin { Any Sector left ? }
    A:=$100; { Sector 0 flag }
    S:=0;
    While (F AND A)=0 do begin { While Sector allready used }
     A:=A SHL 1; { next Sector flag }
     INC(S);     { next Sector }
    end;
    ATrack :=T;
    ASector:=S;
    Dec(BAM.SAT[T],A+1); { Clear Sector Flag & Dec Number of tracks left }
    Dec(FreeBlocks);
    AllocSector:=True;
    exit;
   end;
  end;
  AllocSector:=False;
 end;

function OpenDevice:boolean;
 var
  magic:longint;

  function ReadBAM:boolean;
   var
    i:byte;
   begin
    if (ReadFile(Handle,Magic,4)=4) and (Magic=$64411543) then
     HeaderSize:=64
    else
     HeaderSize:=0;

    if ReadSector(18,0,BAM) then begin
     ReadBAM:=True;
     for i:=1 to 18 do if BAM.Name[i]=#$A0 then BAM.Name[i]:=' ';
     FreeBlocks:=0;
     for i:=1 to 35 do inc(FreeBlocks,BAM.SAT[i] and $FF);
    end else
     ReadBAM:=False;
   end;

 begin
  if Handle=0 then begin
   if D64File='' then SelectFile('Select a D64 File','D64',D64File);
   if D64File='' then Handle:=0 else Handle:=OpenFile(D64File);
   if (Handle=0)or(FileSize(Handle)<683*256)or(not ReadBAM) then begin
    CloseFile(Handle);
    Handle :=0;
   end;
  end;
  OpenDevice:=(Handle>0);
 end;

function LoadDirectory:byte;
 var
  e:word;
  i:byte;
  q:char;

  FirstLine:record

   NextTrack,NextSector:byte;

   ADDR :word;
   Link :word;
   Drive:word;
   RVSOn:byte;
   Quot1:char;
   Name :array[1..18] of char;
   Quot2:char;
   ID   :word;
   Spc1 :char;
   CHARS:word;
   ZERO :byte;

  end absolute Sector;

 begin
  { Directory title }
  with FirstLine do begin
   ADDR :=$0401;
   Link :=$0101;
   Drive:=$0000;
   RVSOn:=$12;
   Quot1:='"';
   Move(BAM.Name,Name,SizeOf(Name));
   Quot2:='"';
   ID   :=BAM.Id;
   Spc1 :=' ';
   CHARS:=BAM.Chars;
   ZERO :=0;
  end;

  SectorSize:=SizeOf(FirstLine)-2;
  SectorIndex:=0;

  Dir.NextTrack :=BAM.Track;
  Dir.NextSector:=BAM.Sector;

  LoadDirectory:=ST_OK;
  FillSector:=fsDir;
 end;

Procedure NextDirectory;
 var
  e,i:word;
  q:char;

  NewLine:record
   Link  :word;
   Blocks:word;
   Space1:char;
   Space2:char;
   Space3:char;
  end;

  EntryLine:record
   Quote1:char;
   Name  :array[1..16] of char;
   Quote2:char;
   Flag1 :char;
   Kind  :array[1..3] of char;
   Flag2 :char;
   ZERO  :byte;
  end;

  LastLine:record
   Sector,Track:byte;
   Link:word;
   Free:word;
   Text:array[1..12] of char;
   ZERO1:word;
   ZERO2:word;
  end absolute Sector;

 begin
  SectorSize :=0;
  SectorIndex:=0;

  While(SectorSize=0)
   and (Dir.NextTrack>0)
   and (ReadSector(Dir.NextTrack,Dir.NextSector,Dir)) do begin
    { Scan all 8 entries of a block }
    for e:=0 to 7 do with Dir.Entry[e] do if Kind<>0 then begin

     NewLine.Link  :=$0101;
     NewLine.Blocks:=Blocks;
     NewLine.Space1:=' ';
     NewLine.Space2:=' ';
     NewLine.Space3:=' ';

     i:=SizeOf(NewLine);
     if Blocks>=10  then dec(i);
     if Blocks>=100 then dec(i);

     Move(NewLine,Sector.Data[SectorSize],i);
     inc(SectorSize,i);

     EntryLine.Quote1:='"'; q:='"';
     for i:=1 to SizeOf(Name) do
      if Name[i]='' then begin Name[i]:=q; q:=' '; end;

     Move(Name,EntryLine.Name,SizeOf(Name));
     EntryLine.Quote2:=q;

     if Kind and $80=0 then EntryLine.Flag1:='*' else EntryLine.Flag1:=' ';
     Move(KindName[Kind and $F],EntryLine.Kind,3);
     if Kind and $40=0 then EntryLine.Flag2:=' ' else EntryLine.Flag2:='<';

     EntryLine.Zero  :=0;

     Move(EntryLine,Sector.Data[SectorSize],SizeOf(EntryLine));
     Inc(SectorSize,SizeOf(EntryLine));

    end; { for entry }

  end;

  if SectorSize=0 then begin
   with LastLine do begin
    Link:=$0101;
    Free:=FreeBlocks;
    Text:='BLOCKS FREE.';
    ZERO1:=0;
    ZERO2:=0;
   end;
   SectorSize:=SizeOf(LastLine)-2;
   FillSector:=fsDone;
  end;

 end;


Function FindFile(Var ATrack,ASector:byte;Exact:boolean):boolean;
 var
{  Dir:TDirectory; }
    n:integer;
  found:boolean;
 begin
  Dir.NextTrack :=BAM.Track;
  Dir.NextSector:=BAM.Sector;

  DirTrack :=BAM.Track;
  DirSector:=BAM.Sector;

  while(Dir.NextTrack<>0)
   and (ReadSector(Dir.NextTrack,Dir.NextSector,Dir)) do begin
    { Scan all 8 entries of a block }
    for DirEntry:=0 to 7 do with Dir.Entry[DirEntry] do if Kind>0 then begin
     found:=true;
     If Exact then begin { SAVE }
      for n:=1 to Length(FileName) do
       if FileName[n]<>Name[n] then found:=false;
      for n:=Length(FileName)+1 to 16 do
       if Name[n]<>'' then found:=false;
     end else begin { LOAD }
      While Length(FileName)<16 do FileName:=FileName+'';
      for n:=1 to Length(FileName) do begin
       if FileName[n]='*' then break; { Found }
       if FileName[n]<>Name[n] then begin
        found:=false;
        break;
       end;
      end;
     {
      if FileName<>'*' then
       for n:=1 to Length(FileName) do if FileName[n]<>Name[n] then found:=false;
     }
     end;
     if found then begin
      ATrack  :=nTrack;
      ASector :=nSector;
      FindFile:=True;
      exit;
     end;
    end;
    DirTrack :=Dir.NextTrack;
    DirSector:=Dir.NextSector;
   end;
  Findfile:=False;
 end;

Function ReadNextSector:byte;
 begin
  if ReadSector(Sector.NextTrack,Sector.NextSector,Sector) then begin
   if Sector.NextTrack=0 then
    SectorSize:=Sector.NextSector-1
   else
    SectorSize:=254;
   SectorIndex:=0;
   if Sector.NextTrack=0 then FillSector:=fsDone else FillSector :=fsRead;
   ReadNextSector:=ST_OK;
  end else begin
   SectorSize:=0;
   FillSector:=fsNone;
   ReadNextSector:=ST_FILENOTFOUND;
  end;
 end;

function LoadFile:byte;
 var
  p:byte;
 begin
  { Convert FileName !!! }
  p:=pos(':',FileName);
  if p>0 then begin
   delete(FileName,1,p);
  end;

  p:=pos(',',FileName);
  if p>0 then begin
   FileName[0]:=chr(p-1);
  end;

  { Now go on }
  if FindFile(Sector.NextTrack,Sector.NextSector,False) then
   LoadFile:=ReadNextSector
  else begin
   StatusBar('['+FileName+'] not found');
   GetKey;
   LoadFile:=ST_TIMEOUT;
   SectorSize :=0;
   SectorIndex:=0;
   FillSector :=fsNone;
  end;
 end;

Function NewFile(Var ATrack,ASector:byte):byte;
 var
  n:byte;
  found:boolean;
 begin
  Dir.NextTrack :=BAM.Track;
  Dir.NextSector:=BAM.Sector;

  DirTrack :=BAM.Track;
  DirSector:=BAM.Sector;

  while(Dir.NextTrack<>0)
   and (ReadSector(Dir.NextTrack,Dir.NextSector,Dir)) do begin
    { Scan all 8 entries of a block }
    for DirEntry:=0 to 7 do with Dir.Entry[DirEntry] do if Kind=0 then begin

     Kind   :=$82;
     AllocSector(nTrack,nSector);
     For n:=1 to 16 do
      if n<=Length(FileName) then Name[n]:=FileName[n] else Name[n]:='';
     sTrack :=0;
     sSector:=0;
     Len    :=0;
     pad0   :=0;
     oTrack :=0;
     oSector:=0;
     Blocks :=0;
     pad1   :=0;

     ATrack :=nTrack;
     ASector:=nSector;

     NewFile:=ST_OK;
     exit;

    end;

    if DirTrack>0 then begin
     DirTrack :=Dir.NextTrack;
     DirSector:=Dir.NextSector;
    end;
   end;

 { Dir.NextTrack=0 }
  AllocSector(Dir.NextTrack,Dir.NextSector);
  WriteSector(DirTrack,DirSector,Dir);

  DirTrack :=Dir.NextTrack;
  DirSector:=Dir.NextSector;

  { New Directory }
  FillChar(Dir,SizeOf(Dir),0);
  Dir.NextSector:=255;
  For DirEntry:=1 downto 7 do begin
   FillChar(Dir.Entry[DirEntry].Name,16,'');
  end;

  DirEntry:=0;
  With Dir.Entry[0] do begin
    Kind   :=$82;
    AllocSector(nTrack,nSector);
    For n:=1 to 16 do
     if n<=Length(FileName) then Name[n]:=FileName[n] else Name[n]:='';

    ATrack :=nTrack;
    ASector:=nSector;
  end;

  NewFile:=ST_OK;
 end;

Procedure TruncFile(Track,Sector:byte);
 Var
  Temp:TSector;
  UnAlloc:Longint;
 begin
  Temp.NextTrack :=Track;
  Temp.NextSector:=Sector;

  While ReadSector(Temp.NextTrack,Temp.NextSector,Temp)
    and(Temp.NextTrack>0) do begin { if there's a Sector After }

  { Unalloc the NEXT Sector }
   UnAlloc:=$100; { split in 2 lines for LongInt SHL }
   UnAlloc:=UnAlloc SHL Temp.NextSector;
   If (BAM.SAT[Temp.NextTrack] and UnAlloc)=0 then begin
    Inc(BAM.SAT[Temp.NextTrack],UnAlloc+1);
    Inc(FreeBlocks);
   end;

  end;
 end;

function SaveFile:byte;
 begin
  if FindFile(Sector.NextTrack,Sector.NextSector,True) then begin

 {  StatusBar('Replace '+FileName+' ? (ENTER/ESC)');
   if GetKey=KEY_ESCAPE then begin
    SaveFile:=ST_NOTPRESENT;
    exit;
   end; }

   TruncFile(Sector.NextTrack,Sector.NextSector);

   Dir.Entry[DirEntry].Blocks:=0; { clear file }

   SaveFile:=ST_OK;

  end else begin

   SaveFile:=NewFile(Sector.NextTrack,Sector.NextSector);

  end;
 end;

function Load(Channel:byte; AFileName:string):byte;
 begin
  if (Handle=0)or(Channel>11) then
   Load:=ST_TIMEOUT
  else begin
   FileName:=AFileName;
   SectorIndex:=0;
   if FileName[1]='$' then Load:=LoadDirectory else Load:=LoadFile;
  end;
 end;

function ReadByte(Channel:byte; Var Value:word):byte;
 begin

  if SectorIndex>=SectorSize then begin
   StatusBar('Reading error');
   ReadByte:=ST_READ_TIMEOUT;
   exit;
  end;

  Value:=Sector.Data[SectorIndex];
  inc(SectorIndex);
  ReadByte:=ST_OK;

  if SectorIndex>=SectorSize then begin

   Case FillSector of
    fsNone : ReadByte:=ST_FILENOTFOUND;
    fsRead : ReadNextSector;
    fsDir  : NextDirectory;
    fsDone : ReadByte:=ST_EOF;
      else   ReadByte:=ST_READ_TIMEOUT;
   end;

  end;

 end;

function WriteByte(Channel:byte; Value,EndOfFile:byte):byte;
 Var
  T,S:byte;
 begin

  if SectorIndex>253 then begin
   WriteByte:=ST_TIMEOUT;
   StatusBar('Write Error !');
   exit;
  end;

  Sector.Data[SectorIndex]:=Value;
  Inc(SectorIndex);
  WriteByte:=ST_OK;

  if (SectorIndex>253)Or(EndOfFile>0) then begin

   T:=Sector.NextTrack;   { current Sector }
   S:=Sector.NextSector;

   if EndOfFile>0 then begin
    Sector.NextTrack :=0;   { last sector }
    Sector.NextSector:=SectorIndex+1;
   end else begin
    AllocSector(Sector.NextTrack,Sector.NextSector); { next sector }
   end;

   WriteSector(T,S,Sector);
   Inc(Dir.Entry[DirEntry].Blocks);

   SectorIndex:=0;

   If EndOfFile>0 then begin
   { Flush }
    WriteSector(18,0,BAM); { for AllocSector }
    WriteSector(DirTrack,DirSector,Dir); { for Dir.Entry[DirEntry].Blocks }
   end;

  end;
 end;


function Save(Channel:byte; AFileName:string):byte;
 begin
  if (Handle=0)or(Channel>11) then
   Save:=ST_TIMEOUT
  else begin
   FileName:=AFileName;
   SectorIndex:=0;
   if FileName='$' then Save:=ST_TIMEOUT else Save:=SaveFile;
  end;
 end;


procedure ChangeDisk;
 begin
  if SelectFile('Select a D64 File','D64',D64File) then begin
   CloseFile(Handle);
   Handle:=0;
   OpenDevice;
  {
   SectorIndex:=255;
   SectorSize :=0;
   FillSector :=fsNone;
  }
  end;
 end;

begin
 Handle :=0;
 D64File:='';
 FillSector:=fsNone;
 SectorIndex:=0;
end.